from torch import nn
import torch.nn.functional as F
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline
import numpy as np
import torch
import pickle


class ResBlock_batchnorm(nn.Module):
    """
    Residual block
    """
    def __init__(self, in_size, channel, out_size):
        """
        Args:
            in_size (int): size of input data vector
            channel (int): size of the inner layer
            out_size (int): size of output data vector

        """
        super(ResBlock_batchnorm, self).__init__()

        self.layer1 = nn.Linear(in_size, channel)
        self.layer2 = nn.Linear(channel, out_size)

        if in_size == out_size:
            self.skip_layer = nn.Identity()
        else:
            self.skip_layer = nn.Linear(in_size, out_size, bias=False)

        self.init_weight()
    def init_weight(self):
        """
        initialize the weight of neural network
        """
        for m in self.modules():
            if type(m)==nn.Linear:
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    m.bias.data.fill_(1E-2)
        nn.init.zeros_(self.skip_layer.weight)

    def forward(self, x):
        """
        Args:
            s (torch tensor): input array
        Returns:
            torch tensor: ourput array

        """
        h = F.relu((self.layer1(x)))
        y = F.relu((self.layer2(h)) * 0.1 + self.skip_layer(x))

        return y


class ChtoModelv2(nn.Module):
    """
        main neural network used by linna
    """
    def __init__(self, in_size, out_size, linearmodel, docpu=False): 
        """
    
        Args:
            in_size (int): size of input data vector
            out_size (int): size of output data vector
            linear model (function): input: tensor array, output: tensor array. One might want to add a pretrained model on top of the NN. 
            docpu (bool) whether to use cpu for evaluation
        """ 
        super(ChtoModelv2, self).__init__()
        self.channel = 16
        hidden_size = max(32, int(out_size*32))
        if out_size>30:
            hidden_size=1000
        self.layer1 = nn.Linear(in_size, hidden_size)
        self.layer2 = ResBlock_batchnorm(hidden_size, self.channel, hidden_size//2) 
        hidden_size = hidden_size//2
        self.layer3 = ResBlock_batchnorm(hidden_size, int(self.channel*2), hidden_size//2) 
        hidden_size = hidden_size//2
        self.layer4 = ResBlock_batchnorm(hidden_size, int(self.channel*4), hidden_size//2) 
        hidden_size = hidden_size//2
        self.layer6 = nn.Linear(hidden_size, hidden_size*4) 
        self.layer7 = nn.Linear(hidden_size*4, out_size) 
        self.layer8 = nn.Linear(out_size, out_size) 
        self.init_weight()
        self.linearmodel = linearmodel
        self.docpu = docpu

    def init_weight(self):
        """
            initialize weights for neural network
        """
        for m in self.modules():
           if type(m)==nn.Linear:
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    m.bias.data.fill_(1E-2)
           elif type(m)==ResBlock_batchnorm:
                m.init_weight()
           elif type(m)==ChtoModelv2:
                pass
           elif type(m)==nn.modules.batchnorm.BatchNorm1d:
                pass
           else:
               print(type(m), flush=True)
               assert(0)
            
    def forward(self, s):
        """

        Args:
            s (torch tensor): input array
        Returns:
            torch tensor: ourput array

        """
        if self.docpu:
            s = s.to_mkldnn()
        s_in  = F.relu(self.layer1(s)) 
        s_in  = self.layer2(s_in) 
        s_in  = self.layer3(s_in) 
        s_in  = self.layer4(s_in) 
        s_in  = F.relu(self.layer6(s_in))
        s_in  = F.relu(self.layer7(s_in))
        if self.linearmodel is not None:
            s = self.layer8(s_in)+self.linearmodel(s)
        else:
            s = self.layer8(s_in)
        if self.docpu:
           s = s.to_dense()
        return s


class ChtoModelv2_linear(nn.Module):
    """
    For testing 
    """
    def __init__(self, in_size, out_size, linearmodel, docpu=False): 
        super(ChtoModelv2_linear, self).__init__()
        self.channel = 16
        hidden_size = max(32, int(out_size*32))
        if out_size>30:
            hidden_size=1000
        self.layer1 = nn.Linear(in_size, hidden_size)
        self.layer2 = ResBlock_batchnorm(hidden_size, self.channel, hidden_size//2) 
        hidden_size = hidden_size//2
        self.layer3 = ResBlock_batchnorm(hidden_size, int(self.channel*2), hidden_size//2) 
        hidden_size = hidden_size//2
        self.layer4 = ResBlock_batchnorm(hidden_size, int(self.channel*4), hidden_size//2) 
        hidden_size = hidden_size//2
        #self.layer5 = ResBlock_batchnorm(hidden_size, self.channel*8, hidden_size//2) 
        #hidden_size = hidden_size//2
        #self.layer6 = ResBlock_batchnorm(hidden_size, self.channel*8, hidden_size//2) 
        #hidden_size = hidden_size//2
        self.layer6 = nn.Linear(hidden_size, hidden_size*4) 
        self.layer7 = nn.Linear(hidden_size*4, out_size) 
        self.layer8 = nn.Linear(out_size, out_size) 
        self.linearlayer  = nn.Linear(in_size, out_size) 
        self.init_weight()
        self.linearlayer.bias.data.fill_(0)
        self.linearlayer.weight.data.fill_(1E-5)
        self.linearmodel = linearmodel
        #if not linearmodel.istrained():
        #    raise(ValueError("linear model must be trained before pass it to ChtoModel_v2"))
        self.docpu = docpu

    def init_weight(self):
        for m in self.modules():
           if type(m)==nn.Linear:
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    m.bias.data.fill_(1E-2)
           elif type(m)==ResBlock_batchnorm:
                m.init_weight()
           elif type(m)==ChtoModelv2_linear:
                pass
           elif type(m)==nn.modules.batchnorm.BatchNorm1d:
                pass
           else:
               print(type(m), flush=True)
               assert(0)
            
    def forward(self, s):
        if self.docpu:
            s = s.to_mkldnn()
        s_in  = F.relu(self.layer1(s)) 
        s_in  = self.layer2(s_in) 
        s_in  = self.layer3(s_in) 
        s_in  = self.layer4(s_in) 
        #s_in  = self.layer5(s_in)
        s_in  = F.relu(self.layer6(s_in))
        s_in  = F.relu(self.layer7(s_in))
        s = self.layer8(s_in) + 1E-3*self.linearlayer(s) 
        if self.docpu:
           s = s.to_dense()
        return s

class LinearModel:
    """
    Linear regression model
    """
    def __init__(self, norder, npc, x_transform=None, y_transform=None, y_inverse_transform_data=None):
        self._istrained=False
        self.norder = norder
        self.npc = npc 
        self.model_poly = None
        self.pcs = None
        self.vec = None 
        self.xmean = None
        self.xstd = None
        self.ymean = None
        self.ystd = None
        if x_transform is None:
            self.x_transform = lambda x: x
        else:
            self.x_transform = x_transform
        if y_transform is None:
            self.y_transform = lambda x:x
        else:
            self.y_transform = y_transform
        if y_inverse_transform_data is None:
            self.y_invtransform_data = lambda x:x
        else:
            self.y_invtransform_data = y_inverse_transform_data
    def train(self, train_x, train_y, sample_weight=None):
        #Do PCA
        self.xmean = torch.mean(train_x, axis=0)
        self.ymean = torch.mean(train_y, axis=0)
        self.xstd = torch.std(train_x, axis=0)
        self.ystd = torch.std(train_y, axis=0)
        train_x_norm = (train_x-self.xmean)/self.xstd
        train_y_norm = (train_y-self.ymean)/self.ystd
        try:
            vec,pcs,  _= torch.linalg.svd(train_y_norm.T.matmul(train_y_norm))
        except:
            L = train_y_norm.T.matmul(train_y_norm)
            print(L.mean())
            vec,pcs,  _= torch.linalg.svd(L + 1e-4*L.mean()*torch.rand(L.shape[0],L.shape[1]))
        if self.npc is None:
            self.npc = np.where(pcs/pcs[0]>0.05)[0][-1]+1
        train_y_projected = train_y_norm.matmul(vec)[:, :self.npc]
        self.vec = vec[:,:self.npc].T
        self.pcs = pcs
        #Train linear model
        self.model_poly = pytorchPolynomialLinear(self.norder)
        self.model_poly.fit(train_x_norm, train_y_projected, sample_weight=sample_weight)
        
        self._istrained=True
    def __call__(self,x):
        xnorm = (x-self.xmean)/self.xstd
        if len(x.shape)==1:
            return self.model_poly(xnorm.reshape(1,-1)).matmul(self.vec)*self.ystd+self.ymean
        else:
            return self.model_poly(xnorm).matmul(self.vec)*self.ystd+self.ymean
        
    def istrained(self):
        return self._istrained

    def save(self, outname):
        f = open(outname, 'wb')
        pickle.dump(self,f)
        f.close()

    def predict(self,x):
        return self.y_invtransform_data(self.y_transform(self(self.x_transform(x))))
        
class pytorchPolynomialLinear:
    """
    Implement a polynomial fit using pytorch

    """
    def __init__(self, ndegree):
        self.poly =  PolynomialFeatures(degree=ndegree)

        self.model_pol = Pipeline([('poly', self.poly),
                  ('linear', LinearRegression(fit_intercept=False))])
        self.coef = None
    def fit(self, train_x, train_y, sample_weight=None):
        kwargs = {self.model_pol.steps[-1][0] + '__sample_weight': sample_weight}
        model_pol = self.model_pol.fit(train_x, train_y, **kwargs)
        self.coef = self.model_pol.named_steps['linear'].coef_

    def trainsform(self, tensor):
        if len(tensor.shape)==1:
            inshape=1
        else:
            inshape = tensor.shape[0]
        result = torch.zeros(size=(inshape, self.poly.n_output_features_))
        powers = self.poly.powers_
        powers = torch.tensor(powers)
        for i, power in enumerate(powers):
            result[:, i] = torch.prod(tensor**power,axis=1)
        return result

    def __call__(self, X):
        return torch.matmul(self.trainsform(X),torch.from_numpy(self.coef.T.astype(np.float32)).to('cpu').requires_grad_())





import sys
import os
import pickle
import time
outdir = sys.argv[1]
timein = sys.argv[2]
if not os.path.isdir(outdir):
    os.makedirs(outdir)
jobid = os.getenv('SLURM_JOB_ID')
nodename = os.getenv('SLURMD_NODENAME')
info = {'jobid': jobid, "nodename": nodename}
outfile = os.path.join(outdir, "gpunodeinfo.pkl")
if os.path.isfile(outfile):
    os.remove(outfile)
with open(outfile, 'wb') as f:
    pickle.dump(info, f)
timesleep = timein.split(":")
timesleep = eval(timesleep[0])*3600+eval(timesleep[1])*60+eval(timesleep[2])
time.sleep(timesleep)



import pytest
from linna.sampler import *
import numpy as np
def test_zeus():
    def log_prob(x):
       return - 0.5 * np.sum(ivar * x**2.0)
    nsteps, nwalkers, ndim = 1000, 10, 2
    ivar = 1.0 / np.random.rand(ndim)
    start = np.random.randn(nwalkers,ndim)
    samp = ZeusSampler(log_prob, ndim, nwalkers,  x0=start)     
    outdir = "tests/out/mcmcout_zeus/"
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    samp.sample(None, nsteps, outdir=outdir, progress=False, overwrite=False, ntimes=10, tautol=0.01, incremental=True)


if __name__=="__main__":
    test_zeus()



import torch
import numpy as np
from tqdm.auto import tqdm


class HMCSampler:
    def __init__(self, lnP, x0, m, transform=None, device='cpu'):
        self.lnP = lnP
        self.x0 = x0.to(dtype=torch.float32, device=device)
        self.x = self.x0.clone()
        self.m = m.to(dtype=torch.float32, device=device)
        self.device = device

        if transform is None:
            self.transform = lambda x: x
        else:
            self.transform = transform

    def sample(self, num_samps, num_steps, step_size):
        chain = []

        # Generate samples.
        for i in tqdm(range(num_samps)):
            # Create initial positions and momenta.
            x = self.x.detach().clone().requires_grad_()
            p = torch.randn(x.shape, device=self.device) * torch.sqrt(self.m)

            # Calculate the initial Hamiltonian and gradients.
            lnP = self.lnP(x)
            prevLnP = lnP.detach().clone()
            H_init = (0.5 * torch.sum(torch.square(p) / self.m) - lnP).detach().cpu().numpy()
            grad = torch.autograd.grad(lnP, x)[0]

            # Initial leapfrog step.
            p += 0.5 * grad * step_size
            x = x + (p / self.m) * step_size

            # Update the gradient.
            lnP = self.lnP(x)
            grad = torch.autograd.grad(lnP, x)[0]

            # Continue leapfrog steps.
            for i in range(1, num_steps):
                p += grad * step_size
                x = x + (p / self.m) * step_size

                lnP = self.lnP(x)
                grad = torch.autograd.grad(lnP, x)[0]

            # Final half-step in momentum.
            p += 0.5 * grad * step_size

            # Calculate final Hamiltonian.
            H_prime = (0.5 * torch.sum(torch.square(p) / self.m) - lnP).detach().cpu().numpy()

            # Perform Metropolis-Hastings accept/reject.
            accept_ratio = np.exp(np.minimum(H_init - H_prime, 0))
            accept_prob = min(accept_ratio, 1)
            if np.random.uniform() < accept_prob:
                sample = {'x': self.transform(x).detach().cpu().numpy(), 'lnP': lnP.detach().cpu().numpy(),
                          'accpet_ratio': accept_ratio, 'accept_prob': accept_prob, 'accepted': True}
                self.x = x
            else:
                sample = {'x': self.transform(self.x).detach().cpu().numpy(), 'lnP': prevLnP.cpu().numpy(),
                          'accpet_ratio': accept_ratio, 'accept_prob': accept_prob, 'accepted': False}
            chain.append(sample)

        return chain



import numpy as np
import pyDOE2
import sample_generator as sg
from copy import deepcopy
import os
from linna.predictor_gpu import * 
import glob
import torch
from torch.utils.data import Dataset, DataLoader
import pickle
import sys
from sklearn.cluster import MeanShift, estimate_bandwidth, KMeans
from linna.nn import *
from scipy.special import erf
from scipy.stats import chi2
import io
from torch.utils import mkldnn as mkldnn_utils
try:
    import mpi4py.rc
    mpi4py.rc.initialize = False
except:
    print("no mpi")

if __name__=="__main__":
    if sys.argv[2] == "cuda":
        while(not torch.cuda.is_available()):
            print("no cuda", flush=True)
    outdir = sys.argv[1]
    f = open(outdir+"/model_pickle.pkl", 'rb')
    model = pickle.load(f) 
    f.close()
    f = open(outdir+"/model_args.pkl", 'rb')
    args = pickle.load(f) 
    f.close()
    model(*args)
    f = open(outdir+"/finish.pkl", 'wb')
    pickle.dump([True], f)
    f.close()








import torch
from torch import nn
from linna.nn import *
from tqdm.auto import tqdm
from linna.nnutils import *
from torch_lr_finder import LRFinder
import copy
from torch.utils import mkldnn as mkldnn_utils
import time
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.distributed as dist
from torch.distributed.algorithms.ddp_comm_hooks.debugging_hooks import noop_hook
from torch.distributed.algorithms.ddp_comm_hooks.powerSGD_hook import *
from numpy import diff




class EarlyStopping(object):
    """
    class handles conditions of terminating neural network training 
    """
    def __init__(self, mode='min', min_delta=0, patience=10, nqueue=200,  percentage=False):
        """
        Args:
            mode (string): Modes to determine whether the loss function is decreasing or not. 
            min_delta (float):  The loss is decreasing if the new score is less than best score - best*min_delta/100 (percentage == False) or less than best score - best*min_delta (percentage==True) 
            parience (int): if best score is not updated after patience steps, the nn will stop training
            nqueue (int): number of step to estimate derivative 
            percentage (bool): determine how min_delta is used 
        """
        self.mode = mode
        self.min_delta = min_delta
        self.patience = patience
        self.best = None
        self.best_t = None
        self.num_bad_epochs = 0
        self.is_better = None
        self._init_is_better(mode, min_delta, percentage)
        self.cooling = 0
        self.cooling_weight_decay = 0
        self.queue_t = []
        self.queue_v = []
        self.nqueue = nqueue 

        if patience == 0:
            self.is_better = lambda a, b: True
            self.step = lambda a: False

    def step(self, metrics, metrics_t):
        """
        Args:
             metrics (float): validation score 
             metrics_t (float): training score
        """
        self.queue_t.append(metrics_t)
        self.queue_v.append(metrics)
        if len(self.queue_t)>self.nqueue:
            self.queue_t.pop(0)
        if len(self.queue_v)>self.nqueue:
            self.queue_v.pop(0)
        running_avg_t = torch.median(torch.stack(self.queue_t))
        running_avg_v = np.median(self.queue_v)
        if len(self.queue_t)>2:
            firsthalft = torch.median(torch.stack(self.queue_t[:int(0.5*len(self.queue_t))]))
            secondhalft = torch.median(torch.stack(self.queue_t[int(0.5*len(self.queue_t)):]))
            firsthalfv = np.median(self.queue_v[:int(0.5*len(self.queue_v))])
            secondhalfv = np.median(self.queue_v[int(0.5*len(self.queue_v)):])
        if self.best is None:
            self.best = metrics
            self.best_t = metrics_t
            self.num_bad_epochs = 0 
        
            return 0
        if np.isnan(metrics):
            print("nan metric", flush=True)
            self.num_bad_epochs += 1
            return 0
        if self.is_better(metrics, self.best):
            self.num_bad_epochs = 0
            self.cooling =0 
            self.cooling_weight_decay = 0
            self.best = metrics
            self.best_t = metrics_t
        else:
            self.num_bad_epochs += 1
            if (self.num_bad_epochs>=self.patience*0.9)&(self.num_bad_epochs<self.patience): 
#                if self.is_better(metrics_t, self.best_t):
#                    self.num_bad_epochs-=1
#                    if self.cooling_weight_decay!=0:
#                        if self.cooling_weight_decay>500:
#                            self.cooling_weight_decay=0
#                            return 0
#                        else:
#                            self.cooling_weight_decay+=1
#                            return 0
#                    else:
#                        self.cooling_weight_decay+=1    
#                        return 3
#                else:
                if (self.cooling!=0):
                    if self.cooling>500:
                        self.cooling=0
                        self.num_bad_epochs += 5
                        return 0
                    else:
                        self.num_bad_epochs-=1
                        self.cooling+=1
                        return 0
                    
                else:
                    self.cooling+=1
                    return 1
            if len(self.queue_t)>2 :
                if  (len(self.queue_t)>0.5*self.nqueue) and (secondhalft-firsthalft <0 ) and (secondhalfv-firsthalfv >0):
                        if self.cooling_weight_decay!=0:
                            if self.cooling_weight_decay>1000:
                                self.cooling_weight_decay=0
                                return 0
                            else:
                                self.queue_t = []
                                self.queue_v = []
                                self.cooling_weight_decay+=1

                                if self.cooling_weight_decay%50==0:
                                    return 3
                                else:
                                    return 0
                        else:
                            self.cooling_weight_decay+=1    
                            return 3
        if self.num_bad_epochs >= self.patience:
            return 2
        return 0

    def _init_is_better(self, mode, min_delta, percentage):
        if mode not in {'min', 'max'}:
            raise ValueError('mode ' + mode + ' is unknown!')
        if not percentage:
            if mode == 'min':
                self.is_better = lambda a, best: a < best - min_delta
            if mode == 'max':
                self.is_better = lambda a, best: a > best + min_delta
        else:
            if mode == 'min':
                self.is_better = lambda a, best: a < best - (
                            best * min_delta / 100)
            if mode == 'max':
                self.is_better = lambda a, best: a > best + (
                            best * min_delta / 100)


class Predictor:
    """
    class handle neural network training and prediction
    """
    def __init__(self, in_size=None, out_size=None, model=None, optim=None, X_transform=None,
                 y_transform=None, device='cpu', scheduler=None, outdir=None):
        """
        Args:
            in_size (int): input size of the neural network 
            out_size (int): output size of the neural network 
            model (string): specify the neural network model defined in nn.py
            optim (pytorch optimizer instance or string, optional): default one is AdamW, if "automatic": will use AdamW with automatically determined learning rate 
            X_transform (callable): transformation function of the input, whose output will be fed into neural network 
            y_transform (callable): transformation function of the output, whouse ouput will be used to training neural network
            device (string): cpu or cuda
            scheduler (None): not used for now [reserved for mpi training]
            outdir (string): specify the output directory
        """
        self.in_size = in_size
        self.out_size = out_size
        self.device = device
        self.best_val_loss= float('inf')
        self.outdir = outdir
        if model is not None:
            self.model = model.to(device)
        else:
            self.model = ChtoModel(in_size, out_size).to(device)
                    
        self.scheduler=scheduler

        if optim is not None:
            
            self.optim = optim
        else:
            self.optim = torch.optim.AdamW(self.model.parameters())

        if X_transform is not None:
            self.X_transform = X_transform
        else:
            self.X_transform = lambda x: x

        if y_transform is not None:
            self.y_transform = y_transform
        else:
            self.y_transform = lambda x: x
        self.MKLDNN=False
        self.MKLDNNMODEL=False

    def train(self, dataset, num_epochs, loss_fn, val_dataset=None, val_metric_fn=None, initfrombest=False, pool=None, nocpu=False, rank=0, size=1):
        """
        training the neural network

        Args:
            dataset (pytorch.DataLoader): iterator of the training dataset
            num_epoches (int): maximum number of epch for traning 
            loss_fn (callable): define the loss function 
            val_dataset (pytorch.DataLoader): iterator of the test dataset 
            val_metric_fn (callabel): define the loss function for validation dataset
            initfrombest (bool): if true, the training will start from the previous best fit point 
            pool (not used): reserved for parallel training 
            nocpu (bool): if true: use gpu to train, else: use cpu to traing
            rank (not used): reserved for parallel training
            size (not used): reserved for parallel training

        Returns:
            np.array: training losses in each epoch 
            np.array: validation losses in each epoch
        """
        torch.manual_seed(1234) 
        if self.optim == "automatic":
            if (rank==0) and (not os.path.isfile(self.outdir+"/lr.npy")):
                newmodel = copy.deepcopy(self.model)
                optimizer = torch.optim.AdamW(newmodel.parameters(), lr=1E-4, weight_decay=1E-4)
                lr_finder = LRFinder(newmodel, optimizer, loss_fn, device=self.device)
                lr_finder.range_test(dataset, val_loader=val_dataset, end_lr=5E-3, num_iter=100)
                fig = plt.figure()
                lr_finder.plot(log_lr=True)
                plt.savefig(os.path.join(self.outdir, "lr_tunning.png"))
                plt.close()
                lrs = lr_finder.history["lr"]
                losses = lr_finder.history["loss"]
                min_grad_idx = (np.gradient(np.array(losses))).argmin()
                lr = lrs[min_grad_idx]
                if lr>1E0:
                   lr = lr/1E2
                np.save(self.outdir+"/lr.npy", lr)
            else:
                while(True):
                    try:
                        lr = np.load(self.outdir+"/lr.npy")
                        break
                    except:
                        pass
        lr = lr*size
        if initfrombest:
            ischeckpoint = self.load_checkpoint()
            if not ischeckpoint:
                print("best.pth.tar does not exsit")
        train_losses = []
        val_metrics = []

        if val_metric_fn is None:
            val_metric_fn = loss_fn
        es = EarlyStopping(patience=500)
        if rank==0:
            pbar = tqdm(range(num_epochs))
        else:
            pbar = range(num_epochs)

        old = 0 
        told = 0
        self.model_old = copy.deepcopy(self.model)
        if pool is not None:
            self.model = DDP(self.model_old)
        self.optim = torch.optim.AdamW(params=self.model.parameters(), lr=lr, weight_decay=1E-4)
        for i in pbar:
            data_iter = iter(dataset)

            self.model.train()

            for X, y_target in data_iter:
                X = X.to(self.device)
                y_target = y_target.to(self.device)

                self.optim.zero_grad()
                y_pred = self.model(self.X_transform(X))
                if pool is not None:
                        loss = loss_fn(y_pred, y_target)
                        all_loss = [torch.zeros_like(loss) for _ in range(size)]
                        loss.backward()
                else:
                    loss = loss_fn(y_pred, y_target) 
                    loss.backward()

                self.optim.step()
                train_losses.append(loss.item())
            if val_dataset is not None:
                val_iter = iter(val_dataset)
                self.model.eval()
                modelmkl = copy.deepcopy(self.model)
                if not nocpu:
                    modelmkl = mkldnn_utils.to_mkldnn(modelmkl)

                val_metric = None
                val_count = 0
                for X, y_target in val_iter:
                    X = X.to(self.device)
                    y_target = y_target.to(self.device)

                    with torch.no_grad():
                        y_pred = modelmkl(self.X_transform(X))

                        if val_metric is None:
                            val_metric = np.array([val.item() for val in val_metric_fn(y_pred, y_target)])
                        else:
                            val_metric += np.array([val.item() for val in val_metric_fn(y_pred, y_target)])

                        val_count += 1

                val_metrics.append(val_metric / val_count)
                if rank==0:
                    pbar.set_description('Train/val Loss: {0:.5e}, {1:.5e}    Epoch'.format(loss, val_metrics[-1][0]))
                if self.outdir is not None:
                    is_best = val_metrics[-1][0] < self.best_val_loss
                    if is_best:
                        self.best_val_loss = val_metrics[-1][0]
                if ((np.std(np.array(val_metrics)[-10:,0])<0.01*(np.mean(np.array(val_metrics)[-10:,0])) and (i>=10) and (i<120) and (i%10==0))):
                    print("bad trainning: {0}".format(i), flush=True)
                    for ind, param_group in enumerate(self.optim.param_groups):
                        lr = param_group['lr']
                    self.model_old.init_weight()
                    del self.model
                    if pool is not None:
                        self.model = DDP(self.model_old)
                    else:
                        self.model = self.model_old
                    self.optim = torch.optim.AdamW(params=self.model.parameters(), lr=lr, weight_decay=1E-4)
                    if ((i>10)&(lr>2E-4)):
                        for ind, param_group in enumerate(self.optim.param_groups):
                            lr = param_group['lr']
                            if lr>2E-6:
                                print("learning rate too large: {0}".format(lr), flush=True)
                                self.optim.param_groups[ind]['lr'] = lr/2.



                if (np.isnan(val_metrics[-1][0])) or (val_metrics[-1][0]>1E10) or ((val_metrics[-1][0]-old>5*old) and (i!=0)) or ((loss-told>5*told) and (i!=0)) :
                    #print( (val_metrics[-1][0]-old>5*old), ((loss-told>5*told)), flush=True)
                    for ind, param_group in enumerate(self.optim.param_groups):
                        lr = param_group['lr']
                    #self.model = self.model_old
                    ischeckpoint = self.load_checkpoint(ismpi=size>1)
                        #if pool is not None:
                            #self.model = DDP(self.model)
                            #state = PowerSGDState(process_group=None, matrix_approximation_rank=1)
                            #self.model.register_comm_hook(state, batched_powerSGD_hook)
                    if (not ischeckpoint):
                        self.model_old.init_weight()
                        del self.model
                        if pool is not None:
                            self.model = DDP(self.model_old)
                            #state = PowerSGDState(process_group=None, matrix_approximation_rank=1)
                            #self.model.register_comm_hook(state, batched_powerSGD_hook)
                        else:
                            self.model = self.model_old
                    self.optim = torch.optim.AdamW(params=self.model.parameters(), lr=lr, weight_decay=1E-4)
                    if (np.isnan(val_metrics[-1][0])) or (val_metrics[-1][0]>1E10) or (val_metrics[-1][0]-old>10*old):
                        if (i>10):
                            for ind, param_group in enumerate(self.optim.param_groups):
                                lr = param_group['lr']
                                if lr>2E-6:
                                    print("learning rate too large: {0}".format(lr), flush=True)
                                    self.optim.param_groups[ind]['lr'] = lr/2.
                    if not (np.isnan(val_metrics[-1][0])):
                        if (val_metrics[-1][0]-old>5*old):
                            val_metrics[-1][0] = old
                    #else:
                    #    for param_group in self.optim.param_groups:
                    #        lr = param_group['lr']
                    #        print("learning rate too large: {0}".format(lr), flush=True)
                    #        param_group['lr'] = lr/2.
                else: 
                    criteria = es.step(val_metrics[-1][0], loss)
                    if criteria == 1:
                        for ind, param_group in enumerate(self.optim.param_groups):
                            lr = param_group['lr']
                            wd = param_group['weight_decay']
                            if lr>2E-6:
                                print("\n learning rate too large: {0}\n".format(lr), flush=True)
                                self.optim.param_groups[ind]['lr'] = lr/2.
                                self.optim.param_groups[ind]['weight_decay'] = wd/2
                            else:
                                es.cooling=0
                        
                    if criteria ==2:
                        print("early stop", flush=True)
                        for ind, param_group in enumerate(self.optim.param_groups):
                            lr = param_group['lr']
                            print("learning rate", lr, flush=True)
                        if rank==0:
                            break
                    if criteria ==3:
                        for ind, param_group in enumerate(self.optim.param_groups):
                            wd = param_group['weight_decay']
                            print("\n weight decay too small: {0}\n".format(wd), flush=True)
                            if wd<1E0:
                                self.optim.param_groups[ind]['weight_decay'] = wd*2
                            #else:
                            #    es.cooling_weight_decay=0
                old = val_metrics[-1][0]
                told = loss
                        
            if (self.outdir is not None)&(rank==0):
                if pool is not None:
                    save_checkpoint({'epoch': i + 1,
                               'state_dict': self.model.module.state_dict(),
                               'optim_dict' : self.optim.state_dict(),
                               'mpi_state_dict': self.model.state_dict()
                               },
                               is_best=is_best,
                               checkpoint=self.outdir)
                else:
                    save_checkpoint({'epoch': i + 1,
                               'state_dict': self.model.state_dict(),
                               'optim_dict' : self.optim.state_dict()},
                               is_best=is_best,
                               checkpoint=self.outdir)


                if (i%100)==0: 
                    train_loss = np.array(train_losses)
                    val_metric = np.array(val_metrics)
                    fig, axes = plt.subplots(1,4,figsize=(15,5))
                    axes[0].plot(np.arange(1, len(train_loss)+1), train_loss, label='Training Mean')
                    axes[0].set_yscale('log')
                    axes[0].legend()
                    axes[1].plot(np.arange(1, len(val_metric[:,0])+1), val_metric[:,0], label='Validation loss')
                    axes[1].set_yscale('log')
                    axes[1].legend()
                    axes[2].plot(np.arange(1, len(val_metric[:,0])+1), val_metric[:,1], label='error max')
                    axes[2].set_yscale('log')
                    axes[2].legend()
                    axes[3].plot(np.arange(1, len(val_metric[:,0])+1), val_metric[:,2], label='error median')
                    axes[3].set_yscale('log')
                    axes[3].legend()
                    plt.xlabel('Epoch')
                    plt.ylabel('$\\chi^2 / dof$')
                    plt.legend()
                    plt.savefig(os.path.join(self.outdir, "training_progress.png"))
                    plt.close()



        if val_dataset is not None:
            return np.array(train_losses), np.array(val_metrics)
        else:
            return np.array(train_losses)

    def load_checkpoint(self, ismpi=False):
        """
        internal use only
        """
        if os.path.isfile(os.path.join(self.outdir, "best.pth.tar")):
            load_checkpoint(os.path.join(self.outdir, "best.pth.tar"), self.model, self.optim, device=self.device, ismpi=ismpi)
            return True
        else:
            return False

    def predict(self, X, no_grad=True):
        """
        make model evaluation 

        Args:
            X (torch tensor): parameters 
            no_grad (bool): True: not keep gradient information, Flase: keep gradient information

        Returns:
            torch.tensor: model evaluation 
        """
        self.model.eval()

        if (len(X.shape) == 1):
            X = X.view(1, -1)
            one_input = True
        else:
            one_input = False
        X_new = self.X_transform(X)
        if self.MKLDNN:
            X_new = X_new.to_mkldnn()
            if not self.MKLDNNMODEL:
                self.MKLDNNmodel = torch.jit.optimize_for_inference(torch.jit.script(mkldnn_utils.to_mkldnn(self.model)))
                self.MKLDNNMODEL=True
            
            if no_grad:
                with torch.no_grad():
                    y_pred = self.MKLDNNmodel(X_new)
            else:
                y_pred = self.MKLDNNmodel(X_new)
        else:
            if no_grad:
                with torch.no_grad():
                    y_pred = self.model(X_new)
            else:
                y_pred = self.model(X_new)

        if self.MKLDNN:
            y_pred = y_pred.to_dense()   
        y_pred = self.y_transform(y_pred)

        if one_input:
            y_pred = y_pred.view(-1)
        return y_pred




import numpy as np
import pyDOE2
import sample_generator as sg
from copy import deepcopy
import os
import glob
import pickle
import sys
import emcee
from linna.nn import *
from scipy.special import erf
from scipy.stats import chi2
import io
import gc
import torch
from torch.utils.data import Dataset, DataLoader
from torch.utils import mkldnn as mkldnn_utils
from linna.util import *
import tempfile


def ml_sampler(outdir, theory, priors, data, cov, init, pool, nwalkers, gpunode, omegab2cut=None, nepoch=4500, method="zeus", nbest=None, chisqcut=None, loglikelihoodfunc=None):
    """
    LINNA main function with hyperparameters set to values described in To et al. 2022

    Args:
        outdir (string): output directory 
        theory (function): theory model 
        priors (dict of str: [float, float]): string can be either flat or gauss. If the string is 'flat', [a,b] indicates the lower and upper limits of the prior. If the string is 'gauss', [a,b] indicates the mean and sigma. 
        data (1d array): float array, data vector 
        cov (2d array): float array, covariance matrix
        init (ndarray): initial guess of mcmc,
        pool (mpi pool, optional): a mpi pool instance that can do pool.map(function, iterables). 
        nwalkers (int) number of mcmc walkers
        gpunode (string): name of gpu node
        omegab2cut (list of int): 2 elements containing the lower and upper limits of omegab*h^2
        nepoch (int, optional): maximum number of epoch for the neural network training
        method (string, optional): Samplers. LINNA supports `emcee` and `zeus`(default)
        nbest (int or list of int): number of points to include in the training set per iteration according to the optimizer
        chisqcut (float, optional): cut the training data if there chisq is greater than this value 
        loglikelihoodfunc (callable, optional): function of model, data , inverse of covariance matrix and return the log liklihood value. If None, then use gaussian likelihood 
    Returns:
        nd array: MCMC chain 
        1d array: log probability of MCMC chain
        
    """
    ntrainArr = [10000, 10000, 10000, 10000]
    nvalArr = [500, 500, 500, 500]
    if method=="emcee":
        nkeepArr = [2, 2, 5, 4]
        ntimesArr = [5, 5, 10, 15]
        ntautolArr = [0.03, 0.03, 0.02, 0.01]
        temperatureArr =  [4.0, 2.0, 1.0, 1.0]
        meanshiftArr = [0.2, 0.2, 0.2, 0.2]
        stdshiftArr = [0.15,0.15,0.15,0.15]
    elif method=="zeus":
        nkeepArr = [2, 2, 5, 5]
        ntimesArr = [5, 5, 10, 50]
        ntautolArr = [0.03, 0.03, 0.02, 0.01]
        temperatureArr =  [4.0, 2.0, 1.0, 1.0]
        meanshiftArr = [0.2, 0.2, 0.2, 0.2]
        stdshiftArr = [0.15,0.15,0.15,0.15]
    else:
        raise NotImplementedError(method)
    dolog10index = None
    ypositive = False 
    device = "cuda"
    docuda=False
    tsize=1
    nnmodel_in = ChtoModelv2
    params = {}
    params["trainingoption"] = 1 
    params["num_epochs"] = nepoch
    params["batch_size"] = 500
    return ml_sampler_core(ntrainArr, nvalArr, nkeepArr, ntimesArr, ntautolArr, meanshiftArr, stdshiftArr, outdir, theory, priors, data, cov,  init, pool, nwalkers, device, dolog10index, ypositive, temperatureArr, omegab2cut, docuda, tsize, gpunode, nnmodel_in, params, method, nbest=nbest, chisqcut=chisqcut, loglikelihoodfunc=loglikelihoodfunc) 

def ml_sampler_core(ntrainArr, nvalArr, nkeepArr, ntimesArr, ntautolArr, meanshiftArr, stdshiftArr, outdir, theory, priors, data, cov,  init, pool, nwalkers, device, dolog10index, ypositive, temperatureArr, omegab2cut=None, docuda=False, tsize=1, gpunode=None, nnmodel_in=None, params=None, method="emcee", nbest=None, chisqcut=None, loglikelihoodfunc=None, nsigma=3):
    """
    LINNA main function 

    Args:
        ntrainArr (int array): number of training data per iteration 
        nvalArr (int array): number of validation data per iteration 
        nkeepArr (int array): number of autocorrelation time to be kept 
        ntimesArr (int array): number of autocorrelation time to stop mcmc
        ntautolArr (float array): error limit of autocorrelation time 
        meanshiftArr (float array): limit on mean shift of parameter estimation from the first and second half of the chain 
        stdshiftArr (float array): limit on std shift of parameter estimation from the first and second half of the chain
        outdir (string): output directory 
        theory (function): theory model 
        priors (dict of str: [float, float]): string can be either flat or gauss. If the string is 'flat', [a,b] indicates the lower and upper limits of the prior. If the string is 'gauss', [a,b] indicates the mean and sigma. 
        data (1d array): float array, data vector 
        cov (2d array): float array, covariance matrix
        init (ndarray): initial guess of mcmc,
        pool (object): mpi4py pool instance 
        nwalkers (int) number of mcmc walkers
        device (string): cpu or gpu
        dolog10index (int array): index of parameters to do log10 
        ypositive (bool): whether the data vector is expected to be all positive 
        temperatureArr (float array): temperature parameters for each iteration
        omegab2cut (list of int): 2 elements containing the lower and upper limits of omegab*h^2  
        docuda (bool): whether do gpu for evaluation 
        tsize (int, optional): number of cores for training 
        gpunode (string): name of gpu node
        nnmodel_in (string): instance of neural network model 
        params (dictionary): dictionary of parameters 
        method (string): sampling method 
        nbest (int or list of int): number of points to include in the training set per iteration according to the optimizer
        chisqcut (float, optional): cut the training data if there chisq is greater than this value 
        loglikelihoodfunc (callable): function of model, data , inverse of covariance matrix and return the log liklihood value 
        nsigma (float): the training point in the first iteration will be generated within nsigma of the gaussian prior 
    Returns:
        nd array: MCMC chain 
        1d array: log probability of MCMC chain
        
    """
    ndim = len(init)
    sigma = np.sqrt(np.diag(cov))
    inv_cov = np.linalg.inv(cov)    
    prior_range = []
    for item in priors:
        if item['dist'] == 'flat':
            prior_range.append([item['arg1'], item['arg2']])
        elif item['dist'] == 'gauss':
            prior_range.append([item['arg1']-5*item['arg2'], item['arg1']+5*item['arg2']])
        else:
            print("not implement dist : {0}".format(item['dist']), flush=True)
            assert(0)
    transform = Transform(priors)
    invtransform = invTransform(priors)
    init = invtransform(init) 
    if method=="emcee":
        filename = "chemcee_256.h5"
    elif method == "zeus":
        filename = "zeus_256.h5"
    else:
        raise NotImplementedError(method)
    for i, (nt, nv, nk, ntimes, tautol, temperature, meanshift, stdshift) in enumerate(zip(ntrainArr, nvalArr, nkeepArr, ntimesArr, ntautolArr, temperatureArr, meanshiftArr, stdshiftArr)):
        if isinstance(nbest, list):
            nbest_in = nbest[i]
            if nbest_in <=0:
                nbest_in = None
        else:
            nbest_in = nbest
        if nbest_in is not None:
            tempdir = tempfile.TemporaryDirectory()
            def negloglike(x):
                d = data-theory([-1,x], tempdir)
                return d.dot(inv_cov.dot(d))
        else:
            negloglike=None
        temperature = temperature**2
        print("#"*100)
        print("iteration: {0}".format(i), flush=True)
        print("#"*100)
        outdir_in = os.path.join(outdir, "iter_{0}/".format(i))
        if i==0:
            chain=None
        else:
            chain_name = os.path.join(os.path.join(outdir, "iter_{0}/".format(i-1)), filename[:-3])
            if os.path.isfile(chain_name+".h5"):
                chain_name = chain_name+".h5"
                chain, _temp, _temp2= read_chain_and_cut(chain_name.format(i-1), nk, ntimes, method=method)
            else:
                chain_name = chain_name+".txt"
                chain = np.loadtxt(chain_name)[-100000:,:-1]
        #Generate training 
        ntrain = nt
        nval = nv
        nnsampler = NN_samplerv1(outdir_in, prior_range) 
        if "trainingoption" in params:
            options = params['trainingoption']
        else:
            options = 0
        generate_training_point(theory, nnsampler, pool, outdir_in, ntrain, nval, data, inv_cov, chain, nsigma=nsigma, omegab2cut=omegab2cut, options=options,  negloglike= negloglike, nbest_in=nbest_in, chisqcut=chisqcut)
        chain = None
        del chain 
        if i!=0:
            try:
                del _temp
                del _temp2
            except:
                pass
        gc.collect()
        if (pool is None) or pool.is_master():
            outdir_list = [os.path.join(outdir, "iter_{0}/".format(m)) for m in range(int(i+1))]

            f = open(outdir_list[-1]+"/model_pickle.pkl", 'wb')
            pickle.dump(train_NN, f) 
            f.close()
            f = open(outdir_list[-1]+"/model_args.pkl", 'wb')
            if gpunode is not None:
                docuda=True
            else:
                docuda=torch.cuda.is_available()
            pickle.dump([nnsampler, cov, inv_cov, sigma, outdir_in, outdir_list, data, dolog10index, ypositive, False, 2, temperature, docuda, None, 1, nnmodel_in, params, nbest_in is not None], f) 
            f.close()
            if not os.path.isfile(outdir_list[-1] + "/finish.pkl"): 
                if gpunode == 'automaticgpu':
                    while(True):
                        gpufile = os.path.join(outdir, "gpunodeinfo.pkl")
                        try:
                            if os.path.isfile(gpufile):
                                with open(gpufile, 'rb') as f:
                                    gpuinfo = pickle.load(f)
                                gpunode = gpuinfo["nodename"]
                                break 
                        except:
                            pass
                if gpunode is not None:
                    print("running gpu on {0}".format(gpunode), flush=True)
                    os.system("cat {2}/train_gpu.py | ssh {0} python - {1} {3}".format(gpunode, outdir_list[-1], os.path.dirname(os.path.abspath(__file__)), "cuda"))
                    while(1):
                        if  os.path.isfile(outdir_list[-1] + "/finish.pkl"):
                            break
                else:
                    os.system("python {1}/train_gpu.py {0} {2}".format(outdir_list[-1], os.path.dirname(os.path.abspath(__file__)), "nocuda"))
                    while(1):
                        if  os.path.isfile(outdir_list[-1] + "/finish.pkl"):
                            break

        #Retrieve model  
        model, y_invtransform_data = retrieve_model(outdir_in, len(init), len(data), nnmodel_in)
        if not docuda:
            model.model = model.model.to(memory_format=torch.channels_last)
            model.MKLDNN=True
        #Do MCMC
        if os.path.isfile(os.path.join(outdir_in, filename)):
            continue
        invcov_new = torch.from_numpy(inv_cov.astype(np.float32)).to('cpu').detach().clone().requires_grad_()
        data_new = torch.from_numpy(data.astype(np.float32)).to('cpu').detach().clone().requires_grad_()
        if loglikelihoodfunc is None:
            loglikelihoodfunc = gaussianlogliklihood
        log_prob = Log_prob(data_new, invcov_new, model, y_invtransform_data, transform, temperature, nograd=True, loglikelihoodfunc=loglikelihoodfunc) 
        dlnp = None
        ddlnp = None
        if pool is not None:
            pool.noduplicate=True
        run_mcmc(nnsampler, outdir_in, method, ndim, nwalkers, init, log_prob, dlnp=dlnp, ddlnp=ddlnp, pool=pool, transform=transform, ntimes=ntimes, tautol=tautol, meanshift=meanshift, stdshift=stdshift, nk=nk)
        if pool is not None:
            pool.noduplicate_close() 
    chain_name = os.path.join(os.path.join(outdir, "iter_{0}/".format(len(ntrainArr)-1)), filename[:-3])
    if os.path.isfile(chain_name+".h5"):
        chain_name = chain_name+".h5"
        chain, log_prob_samples_x, reader = read_chain_and_cut(chain_name.format(len(ntrainArr)-1), nk, ntimes, method=method)
        log_prob_samples_x = reader.get_log_prob(discard=0, flat=True, thin=1)
    else:
        chain_name = chain_name+".txt"
        chain = np.loadtxt(chain_name)[-100000:,:-1]
        log_prob_samples_x = np.loadtxt(chain_name)[-100000:,-1]

    #Optional importance sampling 
    if 'nimp' in params.keys():
        if not os.path.isfile(outdir+"/samples_im.npy"):
            chain_name = os.path.join(os.path.join(outdir, "iter_{0}/".format(len(ntrainArr)-1)), filename[:-3])
            if os.path.isfile(chain_name+".h5"):
                chain_name = chain_name+".h5"
                chain, log_prob_samples_x, reader = read_chain_and_cut(chain_name.format(len(ntrainArr)-1), nk, ntimes, method=method, flat=True)
            else:
                chain_name = chain_name+".txt"
                chain = np.loadtxt(chain_name)[-100000:,:-1]
                log_prob_samples_x = np.loadtxt(chain_name)[-100000:,-1]
            select = np.random.randint(0, len(chain), params['nimp'])
            chain = chain[select]
            log_prob_samples_x = log_prob_samples_x[select]
            np.save(outdir+"/samples_im.npy", chain)
            np.save(outdir+"/log_prob_samples_x.npy", log_prob_samples_x)
        else:
            chain = np.load(outdir+"/samples_im.npy")
            log_prob_samples_x = np.load(outdir+"/log_prob_samples_x.npy")
        outimp = os.path.join(outdir, 'imp/')
        nnsampler = NN_samplerv1(outimp, prior_range)
        if not os.path.isdir(outimp):
            os.makedirs(outimp)
        if not os.path.isfile(outdir+"/theory.npy"):
            theory = nnsampler.generate_training_data(zip(range(len(chain)), chain), theory, pool=pool, args=[outimp])
            np.save(outdir+"/theory.npy", theory)
        else:
            theory = np.load(outdir+"/theory.npy")
        logprior=LogPrior(priors)
        log_prob_samples_x = log_prob_samples_x.flatten()
        logp = logp_theory_data(chain, theory, data, inv_cov, logprior)
        w = np.exp(logp-log_prob_samples_x)
        w[np.abs(np.log(w)-np.mean(np.log(w)))>2*np.std(np.log(w))]=0
        w = w/np.sum(w)
        np.save(outdir+"/weight_im.npy", [log_prob_samples_x.flatten(), logp, w])
    return chain, log_prob_samples_x








import numpy as np
from tqdm import tqdm
from emcee.state import State
import emcee
import os
from scipy.optimize import minimize
from functools import wraps
import torch
import zeus
from zeus.autocorr import AutoCorrTime
import copy
import h5py

def stop_criterion(thetaminus, thetaplus, rminus, rplus, cov=None):
    raise NotImplementedError()

def leapfrog(theta, r, grad, epsilon, f, cov=None):
    raise NotImplementedError()

def build_tree(theta, r, grad, v, j, epsilon, f, joint0, cov):
    raise NotImplementedError()

class Functransform:
    def __init__(self, xmap, u, func, derive=0, torchspeed=False):
        self.u = u
        self.xmap = xmap
        self.orig_lnp = func
        self.derive = derive
        self.torchspeed = torchspeed
    def __call__(self, x, *args, **kwargs):
        if not torch.is_tensor(x):
            x = torch.from_numpy(x.astype(np.float32)).to('cpu').clone().requires_grad_()
        xp = torch.from_numpy(self.xmap.astype(np.float32)).to('cpu').clone().requires_grad_()+torch.from_numpy(self.u.astype(np.float32)).to('cpu').clone().requires_grad_() @ x
        
        if self.derive==1:
            if torch.is_tensor(x):
                xp = xp.detach().numpy()
            return self.u.T @ (self.orig_lnp( xp, *args, **kwargs))
        elif self.derive==2:
            if torch.is_tensor(x):
                xp = xp.detach().numpy()
            return self.u.T @ (self.orig_lnp( xp, *args, **kwargs)) @ self.u
        else:
            if not self.torchspeed:
                lnp = self.orig_lnp( xp.detach().numpy(), *args, **kwargs)
                return lnp.item()
            else:
                lnp = self.orig_lnp( xp, *args, **kwargs)
                return lnp.item(), torch.autograd.grad(lnp,x)[0].detach().numpy()

class CombineFunc:
    def __init__(self, f1, f2):
        self.f1 = f1
        self.f2 = f2
    def __call__(self, x, *args, **kwargs):
        return self.f1(x, *args, **kwargs), self.f2(x, *args, **kwargs)


class _hmc_wrapper(object):
    def __init__(self, random, model_derive, cov, epsilon, nsteps=None):
        self.random = random
        self.nsteps = nsteps
        self.epsilon = epsilon
        self.cov = _hmc_matrix(np.asarray(cov))
        self.model_derive = model_derive #grad_log_probability

    def __call__(self, args):
        coords, current_p = args

        # Sample the initial momentum.
        current_q = coords
        q = current_q
        p = current_p

        # Initial leapfrog step.
        grad_log_probability = self.model_derive(coords)
        p = p + 0.5 * self.epsilon * grad_log_probability

        # Continue leapfrog steps.
        for i in range(self.nsteps):
            # First, a full step in position.
            q = q + self.epsilon * self.cov.apply(p)

            # Then a full step in momentum.
            if i < self.nsteps - 1:
                grad_log_probability = self.model_derive(q)
                p = p + self.epsilon * grad_log_probability

        # Finish with a half momentum step to synchronize with the position.
        grad_log_probability = self.model_derive(q)
        p = p + 0.5 * self.epsilon * grad_log_probability


        # Compute the acceptance probability factor.
        #Factor is the kinetic energy term 
        factor = 0.5 * np.dot(current_p, self.cov.apply(current_p))
        factor -= 0.5 * np.dot(p, self.cov.apply(p))
        return q,  factor


class HamiltonianMove(emcee.moves.Move):
    def __init__(self, compute_derivative, nsteps, epsilon, cov):
        self.nsteps = nsteps
        self.epsilon = epsilon
        self.cov = cov
        self.compute_derivative = compute_derivative

    def get_args(self, ensemble):
        # Randomize the stepsize if requested.
        rand = ensemble.random
        try:
            eps = float(self.epsilon)
        except TypeError:
            eps = rand.uniform(self.epsilon[0], self.epsilon[1])

        # Randomize the number of steps.
        try:
            L = int(self.nsteps)
        except TypeError:
            L = rand.randint(self.nsteps[0], self.nsteps[1])

        return eps, L

    def propose(self, model, state):
        # Set up the integrator and sample the initial momenta.
        nwalkers, ndim = state.coords.shape
        integrator = _hmc_wrapper(model.random, self.compute_derivative,self.cov, *(self.get_args(model)))
        momenta = integrator.cov.sample(model.random, nwalkers,
                                        ndim)

        # Integrate the dynamics in parallel.
        newq = np.zeros((nwalkers, ndim))
        newfactor = np.zeros(nwalkers)
        #for i, (coord, mome) in enumerate(zip(state.coords, momenta)):
        test  = list(model.map_fn(integrator, zip(state.coords, momenta)))
        for n in range(nwalkers):
            newq[n], newfactor[n] = test[n]
        # Compute the lnprobs of the proposed position.
        new_log_probs, new_blobs = model.compute_log_prob_fn(newq)

        # Loop over the walkers and update them accordingly.
        lnpdiff = new_log_probs - state.log_prob +  newfactor
        accepted = np.log(model.random.rand(nwalkers)) < lnpdiff
        
        # Update the parameters
        new_state = State(newq, log_prob=new_log_probs, blobs=new_blobs)
        state = self.update(state, new_state, accepted)

        return state, accepted

def find_reasonable_epsilon(theta0, grad0, logp0, f, cov):
    """ 
        Heuristic for choosing an initial value of epsilon 
    """
    epsilon = 1.
    r0 = np.random.normal(0., 1., len(theta0))

    # Figure out what direction we should be moving epsilon.
    _, rprime, gradprime, logpprime = leapfrog(theta0, r0, grad0, epsilon, f, cov)
    # brutal! This trick make sure the step is not huge leading to infinite
    # values of the likelihood. This could also help to make sure theta stays
    # within the prior domain (if any)
    k = 1.
    while np.isinf(logpprime) or np.isinf(gradprime).any():
        k *= 0.5
        _, rprime, _, logpprime = leapfrog(theta0, r0, grad0, epsilon * k, f, cov)

    epsilon = 0.5 * k * epsilon

    # acceptprob = np.exp(logpprime - logp0 - 0.5 * (np.dot(rprime, rprime.T) - np.dot(r0, r0.T)))
    # a = 2. * float((acceptprob > 0.5)) - 1.
    logacceptprob = logpprime-logp0-0.5*(np.dot(rprime, cov.apply(rprime))-np.dot(r0,cov.apply(r0)))
    a = 1. if logacceptprob > np.log(0.5) else -1.
    # Keep moving epsilon in that direction until acceptprob crosses 0.5.
    # while ( (acceptprob ** a) > (2. ** (-a))):
    while a * logacceptprob > -a * np.log(2):
        epsilon = epsilon * (2. ** a)
        _, rprime, _, logpprime = leapfrog(theta0, r0, grad0, epsilon, f, cov)
        # acceptprob = np.exp(logpprime - logp0 - 0.5 * ( np.dot(rprime, rprime.T) - np.dot(r0, r0.T)))
        logacceptprob = logpprime-logp0-0.5*(np.dot(rprime, cov.apply(rprime))-np.dot(r0,cov.apply(r0)))

    print("find_reasonable_epsilon=", epsilon)

    return epsilon

class NUTSMove(emcee.moves.Move):
    #Modified from  Mfouesneau 's awesome Nuts package 
    #https://github.com/mfouesneau/NUTS/blob/master/nuts/nuts.py
    def __init__(self, lnp, compute_derivative, cov, Madapt, x0, nwalkers, delta=0.6, maxheight=np.inf, torchspeed=True):
        self.cov = _hmc_matrix(np.asarray(cov))
        self.lnp = lnp
        self.dlnp = compute_derivative
        f = self.lnp
        self.epsilon = np.zeros(nwalkers)
        for n, x in enumerate(x0):
            logp, grad = f(x)
            self.epsilon[n] = find_reasonable_epsilon(x, grad, logp, f, self.cov)
        # Parameters to the dual averaging algorithm.
        self.gamma = 0.05
        self.t0 = 10
        self.kappa = 0.75
        self.mu = np.log(10. * self.epsilon)

        # Initialize dual averaging algorithm.
        self.epsilonbar = np.ones(nwalkers)
        self.Hbar = np.zeros(nwalkers)
        self.Madapt = Madapt
        self.m = np.ones(nwalkers)
        self.delta = delta
        self.maxheight = maxheight
        self.torchspeed=torchspeed


    def propose(self, model, state):
        # Set up the integrator and sample the initial momenta.
        nwalkers, ndim = state.coords.shape
        r0s = self.cov.sample(model.random, nwalkers,ndim)

        coords = state.coords
        returncoords=np.zeros_like(state.coords) 
        returnlogps = np.zeros_like(state.log_prob) 
        accepteds =  np.zeros_like(state.log_prob)>1
        returngrad= np.zeros_like(state.blobs)
        function_in = Nuts_one_worker(state, r0s, state.coords, self.dlnp, self.epsilon, model.random, self.cov,  self.maxheight, model.log_prob_fn, self.torchspeed)
    
        test = list(model.map_fn(function_in, range(nwalkers)))
        for n in range(nwalkers):
            returncoords[n], returnlogps[n], accepteds[n], alpha, nalpha, returngrad[n] = test[n]
            if (self.m[n] <= self.Madapt):
                eta = 1. / float(self.m[n] + self.t0)
                self.Hbar[n] = (1. - eta) * self.Hbar[n] + eta * (self.delta - alpha / float(nalpha))
                self.epsilon[n] = np.exp(self.mu[n] - np.sqrt(self.m[n]) / self.gamma * self.Hbar[n])
                eta = self.m[n] ** -self.kappa
                self.epsilonbar[n] = np.exp((1. - eta) * np.log(self.epsilonbar[n]) + eta * np.log(self.epsilon[n]))
            elif self.m[n] == self.Madapt+1:
                print("final epsilon={0} for walker:{1} ".format(self.epsilonbar[n], n))
                self.epsilon[n] = self.epsilonbar[n]
            else:
                pass
            self.m[n] += 1


        new_state = State(returncoords, log_prob=returnlogps, blobs=returngrad)
        state = self.update(state, new_state, accepteds)

        return state, accepteds

class Nuts_one_worker(object):
    def __init__(self, state, r0s, coords, dlnp, epsilon, random, cov,  maxheight, log_prob_fn, torchspeed):
        self.state = state
        self.r0s = r0s
        self.coords = coords
        self.dlnp = dlnp
        self.random = random
        self.cov = cov

        # Initialize dual averaging algorithm.
        self.epsilon = epsilon 
        self. maxheight =  maxheight
        self.log_prob_fn = log_prob_fn
        self.torchspeed = torchspeed

    def f(self, x):
            #x=torch.from_numpy(x.astype(np.float32)).to('cpu').clone().requires_grad_()
            lnP, grad = self.log_prob_fn(x)
            #grad = torch.autograd.grad(lnP, x)[0]
            return lnP, grad 
    def __call__(self, n):
            accepted=False
            joint = self.state.log_prob[n] - 0.5 * np.dot(self.r0s[n].T, self.cov.apply(self.r0s[n]))
            coord = self.coords[n]
            grad = self.dlnp(coord)#self.state.blobs[n] 
            r0 = self.r0s[n]
            # initialize the tree
            thetaminus = coord[:]
            thetaplus = coord[:]
            rminus = r0[:]
            rplus = r0[:]
            gradminus = grad[:]
            gradplus = grad[:]

            j = 0  # initial heigth j = 0
            s = 1  # Main loop: will keep going until s == 0.
            returncoord = coord
            returnlogp = self.state.log_prob[n]
            returngrad = self.state.blobs[n]
            epsilon= self.epsilon[n]
            logptree =0
            while (s==1 and j<self.maxheight):
                # Choose a direction. -1 = backwards, 1 = forwards.
                v = int(2 * (self.random.uniform() < 0.5) - 1)
                if (v == -1):  thetaminus, rminus, gradminus, _, _, _, thetaprime, gradprime, logpprime, sprime, alpha, nalpha, logptree2 = build_tree(thetaminus, rminus, gradminus, v, j, epsilon, self.f, joint, self.cov)
                else:
                    _, _, _, thetaplus, rplus, gradplus, thetaprime, gradprime, logpprime, sprime, alpha, nalpha, logptree2 = build_tree(thetaplus, rplus, gradplus, v, j, epsilon, self.f, joint, self.cov)
                logptot = np.logaddexp(logptree, logptree2)
                if (sprime == 1) and (np.log(self.random.uniform()) < logptree2 - logptot):
                    accepted=True
                    returncoord = thetaprime[:]
                    returnlogp = logpprime
                    returngrad = gradprime[:]
                logptree = logptot
                # Decide if it's time to stop.
                s = sprime and stop_criterion(thetaminus, thetaplus, rminus, rplus, self.cov)
                # Increment depth.
                j += 1
            # Do adaptation of epsilon if we're still doing burn-in.
            return returncoord, returnlogp, accepted, alpha, nalpha, returngrad



class _hmc_matrix(object):

    def __init__(self, var):
        self.var = var

    def sample(self, random, *shape):
        return random.randn(*shape) * np.sqrt(self.var)

    def apply(self, x):
        return x/self.var

class Transformbackend(emcee.backends.HDFBackend):
    def __init__(self,
        filename,
        transform, name="mcmc", read_only=False, dtype=None
     ):
        super(Transformbackend, self).__init__(filename, name, read_only, dtype)
        self.transform = transform

    def reset(self,nwalkers, ndim):
        super().reset(nwalkers, ndim)
        with self.open("a") as f:
            g = f[self.name]
            g.create_dataset(
                "chain_transformed",
                (0, nwalkers, ndim),
                maxshape=(None, nwalkers, ndim),
                dtype=self.dtype,
            )
    def grow(self, ngrow, blobs):
        super().grow(ngrow, blobs)
        with self.open("a") as f:
            g = f[self.name]
            ntot = g.attrs["iteration"] + ngrow
            g["chain_transformed"].resize(ntot, axis=0)
    def save_step(self, state, accepted):
        """Save a step to the backend
        Args:
            state (State): The :class:`State` of the ensemble.
            accepted (ndarray): An array of boolean flags indicating whether
                or not the proposal for each walker was accepted.
        """
        self._check(state, accepted)

        with self.open("a") as f:
            g = f[self.name]
            iteration = g.attrs["iteration"]
            g["chain"][iteration, :, :] = state.coords #[self.transform(c) for c in state.coords]
            g["chain_transformed"][iteration, :, :] = [self.transform(c) for c in state.coords]
            g["log_prob"][iteration, :] = state.log_prob
            if state.blobs is not None:
                g["blobs"][iteration, :] = state.blobs
            g["accepted"][:] += accepted

            for i, v in enumerate(state.random_state):
                g.attrs["random_state_{0}".format(i)] = v

            g.attrs["iteration"] = iteration + 1
        
def checkmeanstd(samples, meanshift, stdshift):
    """
    Check convergence of the mcmc samples by comparing mean and standard deviation estimations of the first half of the sample and the second half 
    
    Args:
       samples (ndarray): (nstep, nwalker, nparam) array. 
       meanshift (float): criteria on shifts in mean 
       stdshift ( float): criteria on shifts in standard deviation

    Returns:
        bool: converged or not 
    """
    firsthalf = samples[:int(len(samples)/2)].reshape(-1, samples.shape[-1])
    secondhalf = samples[int(len(samples)/2):].reshape(-1, samples.shape[-1])
    meanshifte = np.max(np.abs(np.mean(firsthalf, axis=0)-np.mean(secondhalf, axis=0))/np.std(secondhalf, axis=0))
    stdshifte = np.max((np.std(firsthalf, axis=0)-np.std(secondhalf, axis=0))/np.std(secondhalf, axis=0))
    return (meanshifte<meanshift)&(stdshifte<stdshift)

class HMCSampler:
    def __init__(self, lnp, dlnp, ddlnp, ndim, nwalkers,  x0=None, m=None, transform=None, torchspeed=False):
        self.lnp = lnp
        self.dlnp = dlnp
        self.ddlnp = ddlnp
        self.transform = transform
        self.x0 = x0
        self.nparams = ndim
        self.nwalkers=nwalkers
        self.sampler=None


        if m is None:
            self.m = np.ones(self.nparams)
        else:
            self.m = m
        self.torchspeed = torchspeed
        self.calc_hess=False

    def calc_hess_mass_mat(self, maxiter=1E5, gtol=1E-5,resamp_x0=True, tensor=True):
        x = self.x0[0]

        #pbar = tqdm(range(nsteps))
        #for i in pbar:
        #    lnp = self.lnp(x)
        #    grad = self.dlnp(x)
        #    x = x + grad * eps
        #    pbar.set_description('log-prob: {}'.format(lnp))

        res = minimize(lambda x: -1*self.lnp(x).detach().numpy(), x, method='Nelder-Mead',
               options={'maxiter':maxiter, 'disp': True, 'gtol':gtol})

        res = minimize(lambda x: -1*self.lnp(x).detach().numpy(), res.x, method='BFGS',
               jac=lambda x: -1*self.dlnp(x), 
               options={'maxiter':maxiter, 'disp': True, 'gtol':gtol})
        print("optimization done")
        print("#"*100)
        print("#"*100)

        hess = []
        lnp = self.lnp(x)
        hess = self.ddlnp(x)

        u, m, _ = np.linalg.svd(-hess)
        s = 1 / m

        self.u = u
        self.m = m
        self.orig_lnp = self.lnp
        self.orig_dlnp = self.dlnp
        self.orig_ddlnp = self.ddlnp
        self.xmap = x
        self.lnp_in = Functransform(self.xmap, self.u, self.orig_lnp, torchspeed=self.torchspeed)
        self.dlnp = Functransform(self.xmap, self.u, self.orig_dlnp, derive=1)
        if not self.torchspeed:
            self.lnp = CombineFunc(self.lnp_in, self.dlnp)
        else:
            self.lnp = self.lnp_in
        self.ddlnp = Functransform(self.xmap, self.u, self.orig_ddlnp, derive=2)
        if self.transform is None:
            self.transform = lambda x: self.xmap + self.u @ x
        else:
            self.orig_transform = self.transform
            self.transform = lambda x: self.orig_transform(self.xmap + self.u @ x)

        if resamp_x0:
            self.x0 = np.random.randn(*self.x0.shape) *0.5*np.sqrt(s)
        self.calc_hess=True

    def sample(self, pool, nsamp, samp_steps, samp_eps, Madapt=1000, outdir="./", progress=False, overwrite=False, ntimes=10, tautol=0.01, method="hmc", incremental=True, meanshift=0.1, stdshift=0.1, nk=2):
        if (not self.calc_hess)&(method=="nuts"):
           self.lnp = CombineFunc(self.lnp, self.dlnp) 
           self.calc_hess=True
        if self.transform is None:
            self.transform = lambda x: x
        x0 = self.x0
        
        if method=="hmc":
            filename = os.path.join(outdir, "chhmc.h5")
        elif method=="emcee":
            filename = os.path.join(outdir, "chemcee_256.h5")
        elif method == "nuts":
            filename = os.path.join(outdir, "chnuts_test6.h5")
        else:
            print("unknown method: ", method)
            sys.exit()
        if os.path.isfile(filename):
            if overwrite:
                if pool.is_master():
                    os.remove(filename)
            else:
                print("init from previous")
                x0=None
        backend = Transformbackend(filename, self.transform)
        if x0 is None:
            x0 = backend.get_last_sample()
            resume = True
        else:
            resume = False
        dtype = None#[("log_prior", float)]
        if not incremental:
            backend=None
        if method=="hmc":
            moves = [(HamiltonianMove(self.dlnp, samp_steps, samp_eps, self.m),1)] 
            self.sampler = emcee.EnsembleSampler(self.nwalkers, self.nparams, self.lnp, pool=pool,  backend=backend, blobs_dtype=dtype, moves=moves)
        elif method=="emcee":
            self.sampler = emcee.EnsembleSampler(self.nwalkers, self.nparams, self.lnp, pool=pool,  backend=backend, blobs_dtype=dtype)
        elif method == "nuts":
            if self.torchspeed:
                kwargs = {"returntorch":True}
            else:
                kwargs = {}
            print("########")
            moves = [(NUTSMove(self.lnp, self.dlnp,self.m, Madapt=Madapt, x0=x0, nwalkers=x0.shape[0], torchspeed=self.torchspeed, maxheight=5),1)] 
            self.sampler = emcee.EnsembleSampler(self.nwalkers, self.nparams, self.lnp, pool=pool,  backend=backend, blobs_dtype=dtype, moves=moves,kwargs=kwargs)
            
        else:
            print("this should never happen")
            
        # We'll track how the average autocorrelation time estimate changes
        index = 0
        autocorr = np.empty(nsamp)

        # This will be useful to testing convergence
        old_tau = np.inf
        print("start",flush=True)
        if not incremental:
            self.sampler.run_mcmc(x0,nsteps=nsamp, progress=progress, skip_initial_state_check=True);
            return np.array([self.transform(c) for c in self.sampler.get_chain(flat=True)])
        else:
            if not resume:
                print("burnin...", flush=True)
                nwalker = x0.shape[0]
                sampler = emcee.EnsembleSampler(self.nwalkers, self.nparams, self.lnp, pool=pool, blobs_dtype=dtype)
                _ = sampler.run_mcmc(x0,nsteps=100, progress=True, skip_initial_state_check=True);
                flat_chain = sampler.get_chain(flat=True)
                log_prob = sampler.get_log_prob(flat=True)
                pos = flat_chain[np.argsort(log_prob)[::-1][:int(50*nwalker)]]
                x0 = pos[np.random.randint(0,len(pos),nwalker),:]
                print("burnin done...", flush=True)
                self.sampler.reset()    
            for sample in self.sampler.sample(x0, iterations=nsamp, progress=progress, skip_initial_state_check=True):
                # Only check convergence every 100 steps
                if self.sampler.iteration % 100:
                    continue

                # Compute the autocorrelation time so far
                # Using tol=0 means that we'll always get an estimate even
                # if it isn't trustworthy
                tau = self.sampler.get_autocorr_time(tol=0)
                autocorr[index] = np.mean(tau)
                index += 1
                
                if np.isnan(np.sum(tau)) and (self.sampler.iteration>10):
                    break
                # Check convergence
                converged = np.all(tau * ntimes< self.sampler.iteration)
                converged &= np.all(np.abs(old_tau - tau) / tau < tautol)
                converged &= checkmeanstd(self.sampler.get_chain()[-int(nk*np.mean(tau)):], meanshift=meanshift, stdshift=stdshift)
                print("max, min tau diff, max tau, ninter: {0}, {1}, {2}, {3}\n".format(np.max(np.abs(old_tau - tau) / tau), np.min(np.abs(old_tau - tau) / tau), np.max(tau), self.sampler.iteration), flush=True)
                if converged:
                    break

                old_tau = tau
            del self.sampler
            self.sampler = None

class ZeusTransformCallback(zeus.callbacks.SaveProgressCallback):
    def __init__(self, filename, ncheck, transform):
        super(ZeusTransformCallback, self).__init__(filename, ncheck)
        self.transform = transform
    
    def __save(self, x, y):
        x = copy.deepcopy(x)
        with h5py.File(self.directory, 'a') as hf:
            hf['samples'].resize((hf['samples'].shape[0] + x.shape[0]), axis = 0)
            hf['samples'][-x.shape[0]:] = x
            hf['chain_transformed'].resize((hf['samples'].shape[0] + x.shape[0]), axis = 0)
            hf['chain_transformed'][-x.shape[0]:] = np.apply_along_axis(self.transform, -1, x)
            hf['logprob'].resize((hf['logprob'].shape[0] + y.shape[0]), axis = 0)
            hf['logprob'][-y.shape[0]:] = y


    def __initialize_and_save(self, x, y):
        x = copy.deepcopy(x)
        with h5py.File(self.directory, 'w') as hf:
            hf.create_dataset('samples', data=x, compression="gzip", chunks=True, maxshape=(None,)+x.shape[1:])
            hf.create_dataset('chain_transformed', data=np.apply_along_axis(self.transform, -1, x), compression="gzip", chunks=True, maxshape=(None,)+x.shape[1:])
            hf.create_dataset('logprob', data=y, compression="gzip", chunks=True, maxshape=(None,)+y.shape[1:]) 
        self.initialised  = True
    def __call__(self, i, x, y):
        """
        Method that calls the callback function.
        Args:
            i (int): Current iteration of the run.
            x (array): Numpy array containing the chain elements up to iteration i for every walker.
            y (array): Numpy array containing the log-probability values of all chain elements up to
                iteration i for every walker.
        Returns:
            True if the criteria are satisfied and sampling terminates or False if the criteria are
                not satisfied and sampling continues.
        
        """
        if i % self.ncheck == 0:
            if self.initialised:
                self.__save(x[i-self.ncheck:], y[i-self.ncheck:])
            else:
                self.__initialize_and_save(x[i-self.ncheck:], y[i-self.ncheck:])

        return None

    def get_last_sample(self):
        with h5py.File(self.directory, "r") as hf:
            samples = np.copy(hf['samples'])
        return samples[-1]

class Zeusbackend:
    def __init__(self, chainname):
        self.name = chainname

    def get_value(self, name, flat=False, thin=1, discard=0):
        with h5py.File(self.name, "r") as hf:
            v = np.copy(hf[name])
        v = v[discard + thin - 1 : len(v) : thin]
        if flat:
            s = list(v.shape[1:])
            s[0] = np.prod(v.shape[:2])
            return v.reshape(s)
        else:
            return v
    def get_log_prob(self, **kwargs):
        return self.get_value("logprob", **kwargs)

    def get_autocorr_time(self, discard=0, thin=1, **kwargs):
        """
        Args:
            c (float, optional): Truncation parameter of automated windowing procedure of Sokal (1989), default is 5.0
            method (str):  Method to use to compute the integrated autocorrelation time. Available options are ``mk`` (Minas Karamanis method), ``dfm (Daniel Forman-Mackey method)``, and ``gw (Goodman-Weary method)``.

        """
        x = self.get_value("samples", discard=discard, thin=thin)        
        return thin * AutoCorrTime(x, **kwargs)


class ZeusconvergenceCallback:
    """
    Modified from Zeus's autoconvergence callback
    Args:
        ncheck (int): The number of steps after which the IAT is estimated and the tests are performed.
            Default is ``ncheck=100``.
        dact (float): Threshold of the rate of change of IAT. Sampling terminates once this threshold is
            reached along with the other criteria. Default is ``dact=0.01``.
        nact (float): Minimum lenght of the chain as a mutiple of the IAT. Sampling terminates once this threshold is
            reached along with the other criteria. Default is ``nact=10``.
        discard (float): Percentage of chain to discard prior to estimating the IAT. Default is ``discard=0.5``.
        trigger (bool): If ``True`` (default) then terminatate sampling once converged, else just monitor statistics.
        method (str): Method to use for the estimation of the IAT. Available options are ``mk`` (Default), ``dfm``, and ``gw``.
        samples (ndarray): (nstep, nwalker, nparam) array. 
        meanshift (float): criteria on shifts in mean 
        stdshift ( float): criteria on shifts in standard deviation
    """

    def __init__(self, ncheck=100, dact=0.01, nact=10, discard=0.5, trigger=True, method='mk', meanshift=0.1, stdshift=0.1, nk=2):
        self.ncheck = ncheck
        self.dact = dact 
        self.nact = nact

        self.discard = discard
        self.trigger = trigger
        self.method = method

        self.estimates = []
        self.old_tau = np.inf

        self.meanshift = meanshift
        self.stdshift = stdshift
        self.nk=nk

    def __call__(self, i, x, y):
        """
        Method that calls the callback function.
        Args:
            i (int): Current iteration of the run.
            x (array): Numpy array containing the chain elements up to iteration i for every walker.
            y (array): Numpy array containing the log-probability values of all chain elements up to
                iteration i for every walker.
        Returns:
            True if the criteria are satisfied and sampling terminates or False if the criteria are
                not satisfied and sampling continues.
        
        """
        converged = False

        if i % self.ncheck == 0:
        
            tau = np.mean(AutoCorrTime(x[int(i * self.discard):], method=self.method))
            self.estimates.append(tau)

            # Check convergence
            converged = tau * self.nact < i
            converged &= np.abs(self.old_tau - tau) / tau < self.dact
            converged &= checkmeanstd(x[-int(self.nk*tau):], self.meanshift, self.stdshift)
            self.old_tau = tau

        if self.trigger:
            return converged
        else:
            return None


class ZeusSampler:
    def __init__(self, lnp, ndim, nwalkers,  x0=None, transform=None):
        self.lnp = lnp
        self.transform = transform
        self.x0 = x0
        self.nparams = ndim
        self.nwalkers=nwalkers
        self.sampler=None

    def sample(self, pool, nsamp, outdir="./", progress=False, overwrite=False, ntimes=10, tautol=0.01, incremental=True, meanshift=0.1, stdshift=0.1, nk=2):
        if self.transform is None:
            self.transform = lambda x: x
        x0 = self.x0
        filename = os.path.join(outdir, "zeus_256.h5")
        if os.path.isfile(filename):
            if overwrite:
                if pool.is_master():
                    os.remove(filename)
            else:
                print("init from previous")
                x0=None
        backend = ZeusTransformCallback(filename, 100, self.transform)
        if x0 is None:
            x0 = backend.get_last_sample()
            resume = True
        else:
            resume = False
        if not incremental:
            backend=None
        self.sampler = zeus.EnsembleSampler(self.nwalkers, self.nparams, self.lnp, pool=pool, maxiter=1E5)#, moves=[zeus.moves.GlobalMove()])
        cb0 = ZeusconvergenceCallback(ncheck=100, dact=tautol, nact=ntimes, discard=0.2, method='dfm', meanshift=meanshift, stdshift=stdshift, nk=nk)

        if not incremental:
            self.sampler.run_mcmc(x0,nsteps=nsamp, progress=progress);
            return np.array([self.transform(c) for c in self.sampler.get_chain(flat=True)])
        else:
            self.sampler.run_mcmc(x0, nsamp, callbacks=[backend, cb0]);
            del self.sampler
            self.sampler = None



import numpy as np
from copy import deepcopy
import os
import sys
path = "/home/users/chto/code/lighthouse/analysis/"
sys.path.append(path+"/../python/")
sys.path.append(path+"/../python/datavector/")
sys.path.append(path+"/../lib/")
import util_chto
import generate_data_vector
import  generate_batch_theory_datavector
from schwimmbad import MPIPool
import glob
import torch
from torch.utils.data import Dataset, DataLoader
import pickle
import sys
from sklearn.cluster import MeanShift, estimate_bandwidth, KMeans
import numpy as np 
import sys
sys.path.append("/home/users/chto/code/lighthouse/python/datavector/")
import util_chto
sys.path.append("/home/users/chto/code/lighthouse/python/")
from pathlib import Path
import run_4x2ptN_wrapper
import fitsio
import cosmolike_libs_real_mpp_cluster 
import seaborn as sns
import pandas as pd
import argparse
import run_cosmolike_4x2ptN
import run_4x2ptN_wrapper
import pytest
import numpy.testing as npt
from tqdm import tqdm
import os
import random
import signal
import time
from mpi4py.futures import MPIPoolExecutor
from mpi4py import MPI
from linna.main import ml_sampler,ml_sampler_core
from linna.util import *
from linna.nn import *
comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()



def get_prior_dic_init(param):
    (varied_params,
     cosmo_min, cosmo_fid, cosmo_max,
    nuisance_min, nuisance_fid, nuisance_max, cosmo_mean, cosmo_sigma, nuisance_mean, nuisance_sigma) = run_cosmolike_4x2ptN.parse_priors_and_ranges(param)
    nsigma=5
    prior_range = []
    init = []
    for ind, param in enumerate(varied_params):
        if param in cosmo_fid.names():
            sigma = getattr(cosmo_sigma,param)
            if sigma!=0:
                mean = getattr(cosmo_mean,param)
                prior_range.append({'param': param,
                                    'dist': 'gauss',
                                    'arg1': mean,
                                    'arg2': sigma})
            else:
                prior_range.append({'param': param,
                    'dist': 'flat',
                    'arg1': getattr(cosmo_min, param),
                    'arg2': getattr(cosmo_max, param)})
                mean = getattr(cosmo_fid, param)
            init.append(mean)
        elif param in nuisance_fid.names():
            ind_in = int(param.split("_")[-1])
            param_in="_".join(param.split("_")[:-1])
            sigma = getattr(nuisance_sigma,param_in)[ind_in]
            if sigma!=0:
                mean = getattr(nuisance_mean,param_in)[ind_in]
                prior_range.append({'param': param,
                                    'dist': 'gauss',
                                    'arg1': mean,
                                    'arg2': sigma})
            else:
                prior_range.append({'param': param,
                    'dist': 'flat',
                    'arg1': getattr(nuisance_min, param_in)[ind_in],
                    'arg2': getattr(nuisance_max, param_in)[ind_in]})
                mean = getattr(nuisance_fid, param_in)[ind_in]
            init.append(mean)
        else:
            print("no param : ", param)
            assert(0)
    return prior_range, np.array(init)


class Model_func:
    def __init__(self, datavector_writer, mask):
        self.datavector_writer = datavector_writer
        self.mask = mask
    def __call__(self, x, outdirs):
        data = x[1]
        data_file = os.path.join(outdirs, "data_{0}".format(x[0]))
        outfile = bytes(data_file, encoding='utf-8')
        if os.path.isfile(outfile):
            print("skip", outfile)
            return np.loadtxt(outfile.decode())[self.mask,1]
        print("work on: {0}".format(x[0]))
        sys.stdout.flush()
        self.datavector_writer(data, outfile)
        try: 
            data = np.loadtxt(outfile.decode())    
            if len(self.mask)>len(data):
                self.mask = self.mask[:len(data)]
            data = data[self.mask, 1]
        except:
            data = np.zeros_like(np.where(self.mask>0)[0])
        if len(data)==0:
            data = np.zeros_like(np.where(self.mask>0)[0])
        return data

def submitgpujob(allargs):
    outdir = allargs['outdir']
    qos = allargs['qos']
    timein = allargs['time'] 
    gpuconstraint = allargs['gpuconstraint']
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    joboutdir = os.path.join(outdir, "out/")
    if not os.path.isdir(joboutdir):
        os.makedirs(joboutdir)
    jobfile = os.path.join(outdir, "gpujob.sh")
    with open(jobfile, 'w') as fh:
        fh.writelines("#!/bin/bash\n")
        fh.writelines("#SBATCH --job-name=gpu\n")
        fh.writelines("#SBATCH --output={0}/out.dat\n".format(joboutdir))
        fh.writelines("#SBATCH --error={0}/err.dat\n".format(joboutdir))
        fh.writelines("#SBATCH --time={0}\n".format(timein))
        #fh.writelines("#SBATCH --mem=4000\n")
        fh.writelines("#SBATCH -n 1\n")
        fh.writelines("#SBATCH --gres gpu:1\n")
        fh.writelines("#SBATCH  --constraint=\"{0}\"\n".format(gpuconstraint))
        fh.writelines("#SBATCH --gpu_cmode=shared\n")
        fh.writelines("#SBATCH -p {0}\n".format(qos))
        fh.writelines("srun python {0} {1} {2}\n".format(os.path.join(os.path.dirname(os.path.abspath(__file__)), "gpuscript.py"), outdir, timein))
    os.system("sbatch %s" %jobfile)
    

def main():
    import time
    start = time.time()
    method = sys.argv[1]
    params = util_chto.chto_yamlload(sys.argv[3], parent_dir=sys.argv[4])         

    outdir = params['outdir']
    if not os.path.isdir(outdir):
        try:
            os.makedirs(outdir)
        except:
            pass
    gpunode  = sys.argv[2]
    device = "cuda"
    np.random.seed(42)
    nwalkers = 128
    if "nwalkers" in params:
        nwalkers = params['nwalkers']
    ntrainArr = params['ntrainArr']
    nvalArr =  params['nvalArr'] 
    nkeepArr =  params['nkeepArr']
    ntimesArr =  params['ntimesArr']
    ntautolArr =  params['ntautolArr']
    temperatureArr =   params['temperatureArr']
    nnmodel_in = eval(params['nnmodel'])

    if "mask_file" not in params:
        params['mask_file'] = os.path.join(outdir, "mask.dat")
        maskfile = params['mask_file']
        if 'DD_vector' in params.keys():
            try:
                DD_vector = np.loadtxt(os.path.join(params['base_dir'],params["DD_vector"]))[:,1]
                DD_cut = params['DD_cut']
            except:
                DD_vector=None
                DD_cut = None
        else:
            DD_vector = None
            DD_cut = None
        if rank==0:
            run_4x2ptN_wrapper.make_mask_4x2ptN(maskfile, params, notinitialized=True, DD_vector=DD_vector, DD_cut=DD_cut)
        else:
            while(1):
                try:
                    if len(np.loadtxt(maskfile))>0:
                        break
                except:
                    pass
    else:
        maskfile = params['mask_file']
    init_cosmolike = cosmolike_libs_real_mpp_cluster.Initlized_cosmolike(bytes(maskfile, encoding='utf-8'), params)
    init_cosmolike.set_cosmolike()
    priors, init = get_prior_dic_init(params)
    if "omegab2cut" in params:
        omegab2cut = params['omegab2cut']
    else:
        omegab2cut = [3,5,0.005,0.039]
    #try:
    pool = chtoPool(comm)
    #except:
    #    print("no MPI", flush=True)
    #    pool = None
    tsize=1
    mask = np.loadtxt(maskfile)[:,1]
    mask = mask>0
    if (pool is None) or pool.is_master():
            datavector_writer = generate_batch_theory_datavector.generate_base_file(params)
            theory = Model_func(datavector_writer, mask)
   
        
    def readcov(covin):
        cov  = np.zeros((int(np.max(covin[:,0]))+1, int(np.max(covin[:,0]))+1))
        for item in covin:
            cov[int(item[0]), int(item[1])] = item[-2]+item[-1]
            cov[int(item[1]), int(item[0])] = item[-2]+item[-1]
        return cov

    cov = readcov(np.loadtxt(params['base_dir']+params['cov_file']))
    if len(mask)!=len(cov):
        print("mask size not the same as cov, mask is fixed to match cov", flush=True)
        if len(mask)>len(cov):
            mask = mask[:len(cov)]
        else:
            masknew = np.zeros(len(cov))>1
            masknew[:len(mask)] = mask
            mask = masknew

    cov = cov[:,mask][mask,:]

    data = np.loadtxt(params['base_dir']+params['data_file'])[mask,1]
    inv_cov = np.linalg.inv(cov)
    sigma = np.sqrt(np.diag(cov))
   
    if pool is not None:
        if not pool.is_master():
            sys.stdout.flush()
            pool.wait()
            print("done", flush=True)
            print(" mpi = {0} done".format(MPI.COMM_WORLD.rank), flush=True)
            e = sys.exc_info()[0]
            print(e)
            print("#################")
            sys.exit(0)
    if pool is not None:
        if pool.is_master():
            if 'automaticgpu' in params:
                params['automaticgpu']['outdir'] = params['outdir']
                submitgpujob(params['automaticgpu'])
                gpunode = 'automaticgpu'
            ml_sampler_core(ntrainArr, nvalArr, nkeepArr, ntimesArr, ntautolArr, params['meanshiftArr'], params['stdshiftArr'], outdir, theory, priors, data, cov,  init, pool, nwalkers, device, dolog10index=[0,1], ypositive=False, temperatureArr=temperatureArr, omegab2cut=omegab2cut, docuda=False, tsize=tsize, gpunode=gpunode, nnmodel_in=nnmodel_in, params=params, method=method)
            end = time.time() 
            print("Runtime of the program is end - start", end-start)
            np.save(outdir+"/time.npy", end-start)
            if 'automaticgpu' in params:
                gpufile = os.path.join(outdir, "gpunodeinfo.pkl")
                with open(gpufile, 'rb') as f:
                    gpuinfo = pickle.load(f)
                jobid = gpuinfo["jobid"]
                os.system("scancel {0}".format(jobid))


    if pool is not None:
        pool.close()
        





if __name__=="__main__":
    main()



import pytest
import numpy as np
from linna.main import ml_sampler,ml_sampler_core
from linna.util import *
from linna.nn import *
import numpy.testing as npt
ndim = 2 
init =  np.random.uniform(size=ndim)
#covariance matrix 
cov = np.diag([0.5, 0.2])
means = np.array([0.1, 1])

priors = []
for i in range(ndim):
    priors.append({
        'param': 'test_{0}'.format(i),
        'dist': 'flat',
        'arg1': -2.,
        'arg2': 2.
    })

def theory(x, outdirs):
    x_new = deepcopy(x[1])
    return x_new
nwalkers = 4 #Number of mcmc walker
pool = None
ntrainArr = [20]
nvalArr = [5]
nkeepArr = [1]
ntimesArr = [2]
ntautolArr = [0.5]
temperatureArr =  [1.0]
meanshiftArr=[100]
stdshiftArr = [100]
params = {}
params["trainingoption"] = 1
params["num_epochs"] = 10
params["batch_size"] = 5
ypositive=False
dolog10index = None
gpunode = None

def testmain():
    outdir = os.path.abspath(os.getcwd())+"/out/2dgaussian_Fulltconn/"
    chain, logprob = ml_sampler_core(ntrainArr, nvalArr, nkeepArr, ntimesArr, ntautolArr, meanshiftArr, stdshiftArr, outdir, theory, priors, means, cov,  init, pool, nwalkers, "cuda", dolog10index, ypositive, temperatureArr, omegab2cut=None, docuda=False, tsize=1, gpunode=None, nnmodel_in=ChtoModelv2, params=params, method="emcee")

def test_reading():
    outdir = os.path.abspath(os.getcwd())+"/test_data/2dgaussian_Fulltconn/"
    chain, logprob = ml_sampler_core(ntrainArr, nvalArr, nkeepArr, ntimesArr, ntautolArr, meanshiftArr, stdshiftArr, outdir, theory, priors, means, cov,  init, pool, nwalkers, "cuda", dolog10index, ypositive, temperatureArr, omegab2cut=None, docuda=False, tsize=1, gpunode=None, nnmodel_in=ChtoModelv2, params=params, method="emcee")
    npt.assert_almost_equal(np.mean(chain),  0.15151080063411168, decimal=5)
    npt.assert_almost_equal(np.std(chain), 0.9633211647095377, decimal=5)




if __name__ == '__main__':
    testmain()
    test_reading()
